/*
  corp.c - generate an infinite amount of SGI corporate BS

  Originally Written by Chris Pirazzi
                        Silicon Graphics Inc.
                        cpirazzi@sgi.com
                        415-390-5291
                        6/91

  vocabulary enhanced and updated through 8/94 by Chris Pirazzi

  (c) 1994
  permission to modify and distribute is granted as long as name of both
  original author and modifier appear in this comment in modified version.

  Software is provided as-is.  No warranty.  No guarantees.  We take no
  liability for tragic loss of life or productivity or anything else
  using this program!  Use at your own risk!  Do not taunt happy fun
  ball!

  happy BSing !
  
todo:
  verbs like is have to have same # in s and obj
  remove soundfiles from spec
  add weights to each sentence and vocab word 
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <ctype.h>
#include <string.h>
#include <stdarg.h>
#include <limits.h>
#include <assert.h>
#include <sys/time.h>

/* Basic Types and Macros ---------------------------------------- */

typedef int bool;
#define TRUE (1==1)
#define FALSE (!TRUE)

#define max(a,b) ( ((a)>(b)) ? (a) : (b))

/* Utility Routines ----------------------------------------------- */

char *argv0;

void error(char *format, ...)
{
  va_list ap;
  
  fprintf( stderr, "%s: ", argv0 );

  va_start(ap, format);
  vfprintf( stderr, format, ap );
  va_end(ap);

  fprintf( stderr, "\n" );
}

/* strsave -------------------------------------------------- */

/*
   String table functions:

   these functions use a hash table to maintain a unique database 
   of strings.

   note that whenever you call strsave with the same string twice,
   even if the addresses of the passed strings are different,
   strsave returns the same address.
*/

#define HASHSIZE 211

static int hash( char *str )
{
  unsigned int ret;
  for( ret=0; *str; ret=(ret*65599) + *str++ );
  return( (int) (ret%HASHSIZE) );
}

typedef struct linkedstring
{
  char                *string;
  struct linkedstring *next;
} *Linkedstring;

static Linkedstring stringhashtable[HASHSIZE];

char *strsave(char *str)
{
  int i = hash(str);
  Linkedstring p = stringhashtable[i];

  while (p)
    {
      if ( !strcmp( p->string, str ) )
	return(p->string);

      p=p->next;
    }

  /* we didn't find the string -- add it to hash table */
  p = malloc( sizeof(*p) );
  assert(p);
  p->string = malloc( strlen(str)+1 );
  strcpy( p->string, str );
  p->next = stringhashtable[i];
  stringhashtable[i] = p;
  
  return(p->string);
}


/* types --------------------------------------------------------- */

#define MAX_ATTR 15
#define MAX_SOUNDFILE 15

typedef struct soundfile
{
  char *name;
  int fd;
} soundfile;

typedef struct word
{
  char *spelling;
  soundfile soundfiles[MAX_SOUNDFILE];
  int nsoundfiles;
  char *attrs[MAX_ATTR];
  int nattrs;
  struct word *next;  
} word;

typedef struct category
{
  char *name;
  word *words;
  int nwords;
  struct category *next;
} category;

typedef struct vocabulary
{
  category *categs;
  int ncategs;
} vocabulary;

typedef struct clause
{
  word *words[20];
  int nwords;
} clause;

typedef enum eltype { el_categ, el_clause, el_conj } eltype;

typedef struct element
{
  eltype type;
  union 
    {
      char *categ;
      clause *clause;
      struct
	{
	  char *spelling;
	  soundfile sf;
	} conj;
    } u;
} element;

typedef struct sentence
{
  element elements[20];
  int nelements;
  struct sentence *next;
} sentence;

typedef struct sentences
{
  sentence *sents;
  int nsents;
} sentences;

bool hasattr( word *w, char *attr )
{
  char *s = strsave(attr);
  int i;

  for(i=0; i < w->nattrs; i++)
    if (w->attrs[i] == s)
      return TRUE;

  return FALSE;
}


/* vocab input file --------------------------------------------------- */

FILE *fpvocab;
vocabulary vocab;

category *getcategory(char *categ)
{
  char *c = strsave(categ);
  category *cat;

  for( cat=vocab.categs; cat; cat=cat->next )
    if (cat->name == c)
      return cat;

  cat = malloc(sizeof(category));
    
  cat->name = c;
  cat->words = NULL;
  cat->nwords = 0;
  cat->next = vocab.categs;  
  vocab.categs = cat;  
  vocab.ncategs++;  

  return cat;
}

void addword(char *categ, word *w)
{
  category *cat = getcategory(categ);

  w->next = cat->words;
  cat->words = w;
  cat->nwords++;
}

checkword(char *categname, word *w)
{
  category *categ = getcategory(categname);
  int i;

  if (categ->name == strsave("a") ||
      categ->name == strsave("an") ||
      categ->name == strsave("thee") ||
      categ->name == strsave("thuh"))
    {
      /* no problem, these are fixed */
    }
  else if (categ->name == strsave("noun"))
    {
      if (!hasattr(w,"sing")&&!hasattr(w,"plural"))
        error("noun [%s]: no number (sing, plural)", w->spelling);

      if (!hasattr(w,"cons")&&!hasattr(w,"vowel"))
        error("noun [%s]: no cons/vowel", w->spelling);

      if (!hasattr(w,"a")&&!hasattr(w,"the")&&!hasattr(w,"artopt"))
        error("noun [%s]: no a/the/artopt", w->spelling);

      for(i=0; i < w->nattrs; i++)
        {
          char *a = w->attrs[i];
          if (a != strsave("sing") &&
              a != strsave("plural") &&
              a != strsave("cons") &&
              a != strsave("vowel") &&
              a != strsave("a") &&
              a != strsave("the") &&
              a != strsave("artopt"))
            error("noun [%s]: unknown keyword [%s]", 
                  w->spelling, a);
        }
    }
  else if (categ->name == strsave("adj"))
    {
      if (!hasattr(w,"cons")&&!hasattr(w,"vowel"))
        error("adj [%s]: no cons/vowel", w->spelling);

      for(i=0; i < w->nattrs; i++)
        {
          char *a = w->attrs[i];
          if (a != strsave("cons") &&
              a != strsave("vowel"))
            error("adj [%s]: unknown attribute [%s]", 
                  w->spelling, a);
        }
    }
  else if (categ->name == strsave("verb"))
    {
      if (!hasattr(w,"3ps")&&!hasattr(w,"n3ps")&&!hasattr(w,"inf"))
        error("verb [%s]: no 3ps/n3ps/inf", w->spelling);

      if (!hasattr(w,"trans")&&!hasattr(w,"intrans")&&!hasattr(w,"help"))
        error("verb [%s]: no trans/intrans/help", w->spelling);

      if (hasattr(w,"3ps")&&hasattr(w,"inf"))
        error("verb [%s]: suspicious: has 3ps and inf", w->spelling);

      for(i=0; i < w->nattrs; i++)
        {
          char *a = w->attrs[i];
          if (a != strsave("3ps") &&
              a != strsave("n3ps") &&
              a != strsave("inf") &&
              a != strsave("trans") &&
              a != strsave("intrans") &&
              a != strsave("help"))
            error("verb [%s]: unknown keyword [%s]", 
                  w->spelling, a);
        }
      
    }
  else if (categ->name == strsave("adv"))
    {
      if (w->nattrs > 0)
        error("adv [%s]: adverbs have no attributes",
              w->spelling);
    }
  else if (categ->name == strsave("interj"))
    {
      if (w->nattrs > 0)
        error("interj [%s]: interjections have no attributes",
              w->spelling);
    }
  else 
    {
      error("bad category [%s] for word [%s]",
            categ->name, w->spelling);
    }
}

void parsevocabfile( void )
{
  char buf[128];
  
  vocab.categs = NULL;
  vocab.ncategs = 0;  

  while (!feof(fpvocab))
    {
      char *spelling, *sndfile, *categ, *attrs, *p;
      word *wrd;
      int i;
      const char *white = "\t ,\n";

      if (!fgets( buf, 128, fpvocab ))
	continue;

      if (buf[0] == '#')
	continue;

      /* spelling:soundfile,soundfile...:categ:attr,attr,attr... */

      spelling = buf + strspn(buf, white);

      if (!*spelling)
	continue;

      if ( NULL != (sndfile = strchr(spelling, ':')))
	{
	  *sndfile++ = 0;

	  if ( NULL != (categ = strchr(sndfile,':')))
	    {
	      *categ++ = 0;
	      categ += strspn(categ, white);
	      
	      if ( NULL != (attrs = strchr(categ,':')))
		*attrs++ = 0;
	      
	      if ( NULL != (p = strpbrk(categ, white)))
		*p = 0;
	    }	  
	}

      wrd = malloc(sizeof(word));

      wrd->spelling = strsave(spelling);

      wrd->nsoundfiles = 0;
      while (sndfile && *sndfile)
	{
	  sndfile += strspn(sndfile, white);
	  if (!*sndfile) break;
	  if ( NULL != (p = strpbrk(sndfile, white)))
	    *p++ = 0;
	  wrd->soundfiles[wrd->nsoundfiles  ].name = strsave(sndfile);
	  wrd->soundfiles[wrd->nsoundfiles++].fd = -1;
	  sndfile = p;
	}

      wrd->nattrs = 0;
      while (attrs && *attrs)
	{
	  attrs += strspn(attrs, white);
	  if (!*attrs) break;
	  if ( NULL != (p = strpbrk(attrs, white)))
	    *p++ = 0;
	  wrd->attrs[wrd->nattrs++] = strsave(attrs);
	  attrs = p;
	}

      if (!categ)
        fprintf(stderr, "error: no category for %s\n", spelling);

      addword(categ, wrd);
      checkword(categ, wrd);
    }
}

void printvocabulary(void)
{
  category *cat;
  word *w;

  for( cat = vocab.categs; cat; cat=cat->next )
    {     
      printf("CATEGORY [%s] (%d words):\n", cat->name, cat->nwords);

      for( w=cat->words; w; w=w->next )
	{
	  int i;
	  printf("\tWORD [%s] ", w->spelling, w->nsoundfiles);
	  for(i=0; i < w->nattrs; i++ )
	    printf("(%s) ", w->attrs[i]);
	  printf("\n");
	}
    }
}

/* sentence input file ----------------------------------------------- */

FILE *fpsents;

sentences sents;

void killtrailspaces(char *s)
{
  int i = strlen(s) - 1;
  char *white = " \t\n";

  while (i > 0 && strchr(white, s[i]))
    {
      s[i] = 0;
      i--;
    }  
}

void parsesentsfile()
{
  char buf[128];

  sents.sents = NULL;
  sents.nsents = 0;

  while (!feof(fpsents))
    {
      sentence *s;
      char *white = " \t\n";
      char *p, *b, *tmp;

      if (! fgets( buf, 128, fpsents ) )
	continue;

      if (buf[0] == '#')
	continue;

      b = buf + strspn(buf, white);
      if (!*b) continue;
      
      /* text <categ> text <categ> text */
      
      s = malloc(sizeof(sentence));
      s->nelements = 0;
      s->next = sents.sents;
      sents.sents = s;
      sents.nsents++;
      
      while (b)
	{
	  /* get some text (el_conj) */

	  b += strspn( b, white );
	  if (!*b) break;

	  if ( NULL != (p = strchr(b, '<')))
	    *p++ = 0;

	  if (*b)
	    {
	      char *sndfile;

	      if ( NULL != (sndfile = strchr(b, ':')))
		{
		  *sndfile++ = 0;
		  sndfile += strspn(sndfile, white);
		  killtrailspaces(sndfile);
		}
	      
	      killtrailspaces(b);	      

	      s->elements[s->nelements  ].type = el_conj;
	      s->elements[s->nelements  ].u.conj.spelling = strsave(b);
	      if (sndfile)
		s->elements[s->nelements  ].u.conj.sf.name = strsave(sndfile);
	      else
		s->elements[s->nelements  ].u.conj.sf.name = NULL;
	      s->elements[s->nelements++].u.conj.sf.fd = -1;
	    }
	
	  if (!p) break;

	  /* get a category */

	  b = p;

	  b += strspn( b, white );
	  if (!*b) break;

	  if ( NULL != (p = strchr(b, '>')))
	    *p++ = 0;

	  killtrailspaces(b);
	  
	  if (*b)
	    {
	      s->elements[s->nelements  ].type = el_categ;
	      s->elements[s->nelements++].u.categ = strsave(b);
	    }

	  b = p;
	}
    }
}

void printsentences(void)
{
  sentence *s = sents.sents;

  printf("SENTENCES: %d\n", sents.nsents);

  while (s)
    {
      int j;

      for( j=0; j < s->nelements; j++ )
	{
	  element *el = &s->elements[j];

	  switch (el->type)
	    {
	    case el_categ:
	      printf("<%s> ", el->u.categ);
	      break;
	    case el_clause:
	      printf("{CLAUSE!} ");
	      break;
	    case el_conj:
	      printf("[%s]", el->u.conj.spelling);
	      
	      if (el->u.conj.sf.name)
		printf(":[%s]", el->u.conj.sf.name );

	      printf(" ");
	      break;
	    }
	}
      
      printf("\n");
      s = s->next;
    }
}

/* sentence construction  ----------------------------------------------- */

word *getword(char *categ)
{
  category *cat = getcategory(categ);
  int i = random() % cat->nwords;
  word *w = cat->words;

  while (i--) w=w->next;
  
  return w;
}

#define addword(wrd) ( cls->words[cls->nwords++] = wrd )
#define chance(good,total) ( (random() % (total)) < (good) )

word *makenounphrase(clause *cls, bool singular, word *avoid)
{
  word *n, *adj, *afterart;
  bool doart, vowel;

  do { n = getword("noun"); } 
  while (!hasattr(n, singular?"sing":"plural") || n==avoid);

  if (chance(1,2))
    { afterart = n; adj = NULL; }
  else
    afterart = adj = getword("adj"); 
  
  if (hasattr(n, "artopt"))
    {
      if (!hasattr(n,"a") && !hasattr(n,"the"))
	{ doart = FALSE; adj = NULL; }
      else
	doart = chance(1,2);
    }
  else
    doart = TRUE;

  vowel = hasattr(afterart,"vowel");

  if (doart)
    {
      if (!singular || !hasattr(n,"a") || chance(1,2))
	addword(getword(vowel ? "thee" : "thuh"));
      else
	addword(getword(vowel ? "an" : "a"));
    }

  if (adj)
    addword(adj);

  addword(n);

  return n;
}

void makeclause(clause *cls)
{
  word *w, *subj;
  bool subjsing = chance(1,2);
  char *infl, *realinfl;

  cls->nwords = 0;

  subj = makenounphrase(cls, subjsing, NULL);
  
  if (chance(1,70)) 
    addword(getword("interj"));

  infl = subjsing ? "3ps" : "n3ps";

  if (chance(1,10))
    {
      do { w = getword("verb"); } 
      while (!hasattr(w, "help") || !hasattr(w,infl));
      addword(w);	
  
      realinfl = "inf";
    }
  else
    {
      realinfl = infl;
    }
  
  if (chance(1,10))
    addword(getword("adv"));
  
  do { w = getword("verb"); } 
  while (!hasattr(w, realinfl) || hasattr(w, "help"));
  addword(w);

  if (chance(1,50)) 
    addword(getword("interj"));

  if (hasattr(w, "trans"))
    makenounphrase(cls, chance(1,2), subj); 
}

sentence *getsentence(void)
{
  int i = random() % sents.nsents;
  sentence *s = sents.sents;

  while (i--) s=s->next;
  
  return s;  
}

sentence sent;

void makesentence(void)
{  
  sentence *s = getsentence();
  element *elin, *elout;
  int i;

  sent.nelements = s->nelements;
  sent.next = NULL;		/* (no meaning here) */

  for( i=0; i < sent.nelements; i++ )
    {
      elin  = &s->elements[i];
      elout = &sent.elements[i];
      
      if (elin->type == el_clause)
	fprintf(stderr,"oops ... clause in sentence template!\n");
      else if (elin->type == el_conj)
	*elout = *elin;
      else			/* elin->type == el_categ -- choose one! */
	{
	  elout->type = el_clause;
	  elout->u.clause = malloc(sizeof(clause));
	  
	  if (elin->u.categ == strsave("clause"))
	    {
	      makeclause(elout->u.clause);
	    }
	  else if (elin->u.categ == strsave("nphrase"))
	    {
	      elout->u.clause->nwords = 0;
	      makenounphrase(elout->u.clause, chance(1,2), NULL);
	    }
	  else if (elin->u.categ == strsave("pnphrase"))
	    {
	      elout->u.clause->nwords = 0;
	      makenounphrase(elout->u.clause, FALSE, NULL);
	    }
	  else if (elin->u.categ == strsave("snphrase"))
	    {
	      elout->u.clause->nwords = 0;
	      makenounphrase(elout->u.clause, TRUE, NULL);
	    }
	  else
	    {
	      elout->u.clause->nwords = 1;
	      elout->u.clause->words[0] = getword(elin->u.categ);
	    }
	}
      
    }
}


/* sentence performance ----------------------------------------------- */

void nextthing(char *s, bool *didcap, int *curx)
{
  char *white = " \t";
  char *s2 = strdup(s), *p, *b;

  b = s2;

  while (b && *b)
    {
      int slen, spclen;
      bool begin = !*didcap;
      bool punct =
	(b[0]==',' || b[0]==':' || b[0]==';' || (b[0]=='.' && b[1]!='.'));

      if (begin)
	{ b[0] = toupper(b[0]);  *didcap = TRUE; }

      if ( NULL != (p = strpbrk(b, white)))
	*p++ = 0;
      
      slen = strlen(b);

      if (punct)
	spclen = slen-1;	/* punctuation can leak 1 over margin */
      else if (begin) 
	spclen = slen + 2;
      else
	spclen = slen + 1;
	
      if (*curx + spclen > 76)
	{
	  printf("\n");
	  *curx = slen;
	}
      else
	{
	  *curx += spclen;
	  if (!punct)
	    printf(begin ? "  " : " ");
	}
      printf("%s", b);
       
      b = p;      
    }

  free(s2);        
}

void dosentence(int *curx)
{
  int j;
  bool didcap = FALSE;  

  for(j=0; j < sent.nelements; j++)
    {   
      char *s;

      if (sent.elements[j].type == el_categ)
	{
	  fprintf(stderr,"oops ... category not changed "
		  "to clause in sentence\n");
	  continue;
	}

      if (sent.elements[j].type == el_conj)
	nextthing(sent.elements[j].u.conj.spelling, &didcap, curx );
      else
	{
	  int i;
	  clause *cls = sent.elements[j].u.clause;
	  for(i=0; i<cls->nwords; i++)
	    nextthing( cls->words[i]->spelling, &didcap, curx );
	}
    }
}

/* main ------------------------------------------------------------ */

void main( int argc, char **argv )
{
  int i, curx, seed;
  int nparagraphs = -1;
  argv0 = argv[0];

  if (argc < 3)
    {
      printf("%s usage: %s vocabfile sentencefile\n", argv[0], argv[0] );
      exit(0);
    }
  
  if (NULL == (fpvocab = fopen( argv[1], "r" )))
    {
      perror("fopen");
      exit(2);
    }

  if (NULL == (fpsents = fopen( argv[2], "r" )))
    {
      perror("fopen");
      exit(2);
    }

  if ( argc >= 4 && (seed=atoi(argv[3]))!=0 )
    {
      srandom(seed);
    }
  else
    {
      struct timeval tm;
      gettimeofday(&tm);
      srandom(tm.tv_usec);
    }

  if (argc >= 5)
    nparagraphs = atoi(argv[4]);

  parsevocabfile();
  parsesentsfile();

  /* printvocabulary(); */
  /* printsentences(); */ 

  curx = 2;
  printf("  ");

  while (1)
    {
      makesentence();     
      dosentence(&curx);
      /* new paragraph */
      if (chance(1,10))
	{
	  printf("\n\n  ");
          if (nparagraphs >= 0)
            {
              nparagraphs--;
              if (nparagraphs<=0)
                exit(0);
            }
	  curx = 0;
	}
    }
}

